#!/usr/bin/python

# texture = filename, palette
# vertex = x, y, z, u, v, light
# polygon = a, b, c, visibility, texture
#{a = {x=123, y=234, z=231, u=432, v=122, light={p=-2342, q=3422, r=-2102}} b=... vis=[0, 1, 0, 1, 0, 0...] tex=12}

import sys, os, math, struct
import fft.map

class File(file):

    def __init__(self, fileName):
        file.__init__(self, fileName, 'rb')
        self.toc = {}
        for i in range(49):
            self.toc[i*4] = struct.unpack('<I', self.read(4))[0]

    def readShort(self, signed = 1):
        inData = self.read(2)
        if signed:
            inData = struct.unpack("<h", inData)
        else:
            inData = struct.unpack("<H", inData)
        return inData[0]

    def readVertex(self):
        ret = {'u': None, 'v': None, 'p': None, 'q': None, 'r': None}
        (ret['x'], ret['y'], ret['z']) = struct.unpack('<hhh', self.read(6))
        ret['y'] = - ret['y']
        return ret

    def readVertexLight(self):
        ret = {}
        (ret['p'], ret['q'], ret['r']) = struct.unpack('<hhh', self.read(6))
        return ret

    def readUV(self):
        ret = {}
        (ret['u'], ret['v']) = struct.unpack('BB', self.read(2))
        return ret

    def readTriangle(self):
        ret = {}
        ret["a"] = self.readVertex()
        ret["b"] = self.readVertex()
        ret["c"] = self.readVertex()
        return ret

    def readTriangleLight(self):
        ret = {}
        ret["a"] = self.readVertexLight()
        ret["b"] = self.readVertexLight()
        ret["c"] = self.readVertexLight()
        return ret

    def readSquTexture(self):
        ret = {}
        ret["a"] = self.readUV()
        ret["palette"] = ord(self.read(1))
        junk = self.read(1)
        ret["b"] = self.readUV()
        ret["page"] = ord(self.read(1))
        junk = self.read(1)
        ret["c"] = self.readUV()
        ret["d"] = self.readUV()
        return ret

    def getPalettes(self):
        self.seek(self.toc[fft.map.TOC_PALETTES_COLOR])
        for paletteNum in range(16):
            palette = []
            for colorNum in range(16):
                inColor = self.readShort(0)
                t = (inColor & 0x8000) >> 15
                b = (inColor & 0x7C00) >> 10
                g = (inColor & 0x03E0) >> 5
                r = (inColor & 0x001F) >> 0
                palette.append((r,g,b,t))
            yield palette

    def getGLPalettes(self):
        self.seek(self.toc[fft.map.TOC_PALETTES_COLOR])
        palette_set = []
        for paletteNum in range(16):
            palette = []
            for colorNum in range(16):
                inColor = self.readShort(0)
                t = (inColor & 0x8000) >> 15
                b = (inColor & 0x7C00) >> 10
                g = (inColor & 0x03E0) >> 5
                r = (inColor & 0x001F) >> 0
                palette.append(r)
                palette.append(g)
                palette.append(b)
            palette_set.append(palette)
        return palette_set

    def getGLPalettesAlpha(self):
        self.seek(self.toc[fft.map.TOC_PALETTES_COLOR])
        palette_set = []
        for paletteNum in range(16):
            palette = []
            for colorNum in range(16):
                inColor = self.readShort(0)
                t = (inColor & 0x8000) >> 15
                b = (inColor & 0x7C00) >> 10
                g = (inColor & 0x03E0) >> 5
                r = (inColor & 0x001F) >> 0
                if (r == 0 and g == 0 and b == 0 and t == 0):
                    palette.append(0)
                    palette.append(0)
                    palette.append(0)
                else:
                    palette.append(255)
                    palette.append(255)
                    palette.append(255)
            palette_set.append(palette)
        return palette_set

    def getLights(self):
        lights = [{}, {}, {}, {}]
        self.seek(self.toc[fft.map.TOC_LIGHTS_BACKGROUND])
        for color in ["r", "g", "b"]:
            for num in range(3):
                lights[num][color] = self.readShort() * 1.0
        for num in range(3):
            for dim in ["x", "y", "z"]:
                lights[num][dim] = self.readShort()
            lights[num]["y"] = -lights[num]["y"]
        lights[3]["r"] = ord(self.read(1))
        lights[3]["g"] = ord(self.read(1))
        lights[3]["b"] = ord(self.read(1))
        return lights

    def getTerrain(self):
        self.seek(self.toc[fft.map.TOC_TERRAIN])
        max_x = ord(self.read(1))
        max_z = ord(self.read(1))
        terrain = []
        for y in range(2):
            level = []
            for z in range(max_z):
                row = []
                for x in range(max_x):
                    tile = {}
                    short = self.readShort(0)
                    tile["surface"] = short & 0x3f
                    short = self.readShort(0)
                    tile["height"] = short & 0xff
                    tile["depth"] = short >> 13
                    tile["slopeheight"] = ((short >> 8) & 0x1f)
                    short = self.readShort(0)
                    tile["slopedir"] = short & 0xff
                    short = self.readShort(0)
                    tile["nowalk"] = (short >> 1) & 0x01
                    tile["nocursor"] = short & 0x01
                    row.append(tile)
                level.append(row)
            terrain.append(level)
            self.seek(self.toc[fft.map.TOC_TERRAIN] + 2 + 256 * 8)
        return terrain

    def getBGGrad(self):
        grad = [{}, {}]
        self.seek(self.toc[fft.map.TOC_LIGHTS_BACKGROUND] + 39)
        grad[0]["r"] = ord(self.read(1)) * 1.0
        grad[0]["g"] = ord(self.read(1)) * 1.0
        grad[0]["b"] = ord(self.read(1)) * 1.0
        grad[1]["r"] = ord(self.read(1)) * 1.0
        grad[1]["g"] = ord(self.read(1)) * 1.0
        grad[1]["b"] = ord(self.read(1)) * 1.0
        return grad
