import os, math, struct
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.fragment_shader import *
from OpenGL.GL.ARB.vertex_shader import *
import Image

try:
    from OpenGL.extensions import alternate
except ImportError:
    pass
else:
    glCreateShader = alternate( 'glCreateShader', glCreateShader, glCreateShaderObjectARB )
    glShaderSource = alternate( 'glShaderSource', glShaderSource, glShaderSourceARB)
    glCompileShader = alternate( 'glCompileShader', glCompileShader, glCompileShaderARB)
    glCreateProgram = alternate( 'glCreateProgram', glCreateProgram, glCreateProgramObjectARB)
    glAttachShader = alternate( 'glAttachShader', glAttachShader,glAttachObjectARB )
    glValidateProgram = alternate( 'glValidateProgram',glValidateProgram,glValidateProgramARB )
    glLinkProgram = alternate( 'glLinkProgram',glLinkProgram,glLinkProgramARB )
    glDeleteShader = alternate( 'glDeleteShader', glDeleteShader,glDeleteObjectARB )
    glUseProgram = alternate('glUseProgram',glUseProgram,glUseProgramObjectARB )

def compileShader( source, shaderType ):
        """Compile shader source of given type"""
        shader = glCreateShader(shaderType)
        glShaderSource( shader, source )
        glCompileShader( shader )
        return shader

def compileProgram(vertexSource=None, fragmentSource=None):
        program = glCreateProgram()

        if vertexSource:
                vertexShader = compileShader(
                        vertexSource, GL_VERTEX_SHADER_ARB
                )
                glAttachShader(program, vertexShader)
        if fragmentSource:
                fragmentShader = compileShader(
                        fragmentSource, GL_FRAGMENT_SHADER_ARB
                )
                glAttachShader(program, fragmentShader)

        glValidateProgram( program )
        glLinkProgram(program)

        try:
            if vertexShader:
                    glDeleteShader(vertexShader)
            if fragmentShader:
                    glDeleteShader(fragmentShader)
        except:
            pass

        return program

class TexEdit():
    def __init__(self):
        self.textures = []
        self.palette = 0
        self.show_used_areas = True
        self.map = None

    def gl_initialize(self):
        glEnable(GL_BLEND)
        glEnable(GL_ALPHA_TEST)
        glAlphaFunc(GL_GREATER, 0.5)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(0.0, 0.0, 0.0, 0.0)    # Clear the Background to a Color
        glClearDepth(1.0)                   # Enables Clearing Of The Depth Buffer
        glDepthFunc(GL_LEQUAL)                # The Type Of Depth Test To Do
        glEnable(GL_DEPTH_TEST)             # Enables Depth Testing
        glShadeModel(GL_SMOOTH)             # Enables Smooth Color Shading
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        self.gl_create_shaders()
        glUseProgram(self.shader_program)
        self.uniform_texture = glGetUniformLocation(self.shader_program, 'tex')
        self.uniform_palette = glGetUniformLocation(self.shader_program, 'palette')
        glUniform1i(self.uniform_palette, self.palette)
        self.uniform_texpalette = glGetUniformLocation(self.shader_program, 'tex2')

    def gl_set_viewport(self, width, height):
        # The function called when our window is resized
        if height == 0: # Prevent A Divide By Zero If The Window Is Too Small
            height = 1
        # Reset The Current Viewport And Perspective Transformation
        glViewport(0, 0, width, height)
        self.gl_set_perspective()

    def gl_set_perspective(self):
        # Reset The Projection Matrix
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        # Set orthographic or perspective projection
        aspect = 512.0 / 512.0
        self.orthobase = 256.0
        glOrtho(-self.orthobase, self.orthobase, -self.orthobase / aspect,
                self.orthobase / aspect, 0.1, 24000)
        glMatrixMode(GL_MODELVIEW)

    def load_map(self, map, init):
        self.map = map
        if init:
            self.gl_create_textures()

    def gl_create_textures(self):
        if self.map.texture == None:
            glDisable(GL_TEXTURE_2D)
            glColor4f(1.0, 1.0, 1.0, 1.0)
            return
        texFileName = self.map.texture
        # Create Texture
        if self.textures:
            glDeleteTextures(self.textures)
            self.textures = []
        glUseProgram(self.shader_program)
        newtex = glGenTextures(1)
        self.textures.append(newtex)
        glBindTexture(GL_TEXTURE_2D, newtex)
        image = Image.open(texFileName).convert("P")
        ix = image.size[0]
        iy = image.size[1]
        gray_palette = [(x / 3) * 17 for x in range(48)]
        # Palette loading, should probably go elsewhere
        texpalette = Image.new('RGBA', (256, 1));
        for p in range(16):
            for i in range(16):
                rgb = self.map.palettes_color[p][i*3:i*3+3]
                a = self.map.palettes_color_alpha[p][i * 3]
                texpalette.putpixel((p * 16 + i, 0), tuple(rgb + [a]))
        paletteimagestr = texpalette.tostring('raw', 'RGBA', 0, -1)
        newtexpalette = glGenTextures(1)
        self.textures.append(newtexpalette)
        glActiveTexture(GL_TEXTURE1)
        glBindTexture(GL_TEXTURE_1D, newtexpalette)
        glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, paletteimagestr)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glUniform1i(self.uniform_texpalette, 1)
        # End palette loading
        image.putpalette(gray_palette)
        #image.save("tex.png")
        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, newtex)
        imagestr = image.convert('RGBA').tostring("raw", "RGBA", 0, -1)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1, ix, iy, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagestr)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)

    def gl_create_shaders(self):
        self.shader_program = compileProgram(
[
        '''
        void main() {
            gl_TexCoord[0] = gl_MultiTexCoord0;
            gl_Position = ftransform();
        }
        '''
],
[
        '''
        uniform sampler2D tex;
        uniform sampler1D tex2;
        uniform int palette;
        void main() {
            vec4 texel, ct;
            float index, subscript;

            texel = texture2D(tex,gl_TexCoord[0].st);
            index = texel.r * 15.0;
            index = sign(index) * floor(abs(index + 0.5));
            subscript = float(palette) * 16.0 + index;
            ct = texture1D(tex2, subscript / 255.0) * 8.225806;

            gl_FragColor = ct;
        }
        ''',
]
)
        glUseProgram(self.shader_program)

    def gl_draw(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()                    # Reset The View

        glTranslate(-256, -256, -10)

        glBegin(GL_QUADS)
        glTexCoord2f(0, 0)
        glVertex3i(0, 0, 0)
        glTexCoord2f(1.0, 0)
        glVertex3i(256, 0, 0)
        glTexCoord2f(1.0, 0.5)
        glVertex3i(256, 512, 0)
        glTexCoord2f(0, 0.5)
        glVertex3i(0, 512, 0)
        glTexCoord2f(0, 0)
        glVertex3i(0, 0, 0)
        glEnd()

        glBegin(GL_QUADS)
        glTexCoord2f(0, 0.5)
        glVertex3i(256, 0, 0)
        glTexCoord2f(1.0, 0.5)
        glVertex3i(512, 0, 0)
        glTexCoord2f(1.0, 1.0)
        glVertex3i(512, 512, 0)
        glTexCoord2f(0, 1.0)
        glVertex3i(256, 512, 0)
        glTexCoord2f(0, 0.5)
        glVertex3i(256, 0, 0)
        glEnd()

        if self.show_used_areas:
            self.gl_draw_used_areas()

    def gl_draw_used_areas(self):
        if not self.map:
            return
        glUseProgram(0)
        glColor4f(1.0, 1.0, 1.0, 1.0)
        for polygon in self.map.polygons:
            if polygon['palette'] != self.palette:
                continue
            glBegin(GL_LINE_STRIP)
            page = polygon['page']
            for point in ['a', 'b', 'c', 'a']:
                vertex = polygon[point]
                if vertex['u'] == None:
                    continue
                u_offset = (1 - page / 2) * 256
                v_offset = (1 - page % 2) * 256
                glVertex3f(vertex['u'] + u_offset + 1, 256 - vertex['v'] + v_offset, 1.0)
            glEnd()
        glUseProgram(self.shader_program)

    def gl_set_palette(self, index):
        assert 0 <= index < 16
        self.palette = index
        glUniform1i(self.uniform_palette, index)

    def set_show_used(self, state):
        self.show_used_areas = bool(state)

    def export_tex(self, dest_file_name):
        if self.map.texture == None:
            return
        texFileName = self.map.texture
        image = Image.open(texFileName).convert("P")
        bright_palette = []
        factor = 255.0 / 31.0
        for value in self.map.palettes_color[self.palette]:
            bright_palette.append(round(value * factor))
        image.putpalette(bright_palette)
        image.save(dest_file_name)

    def import_tex(self, src_file_name):
        if self.map.texture == None:
            return
        image = Image.open(src_file_name)
        assert image.mode == 'P', 'Input file must be a paletted PNG.'
        gray_palette = []
        for i in range(16):
            gray_palette.extend([i] * 3)
        image.putpalette(gray_palette)
        image_data = list(image.getdata())
        assert len(image_data) == 256 * 1024, 'Input file must be 256x1024 pixels.'
        texFileName = self.map.texture
        outfile = open(texFileName, 'wb')
        for i in range(len(image_data) / 2):
            byte = image_data[i*2]
            byte += image_data[i*2+1] << 4
            byte = struct.pack('B', byte)
            outfile.write(byte)
        outfile.close()
