
import os, math
from OpenGL.GL import *
from OpenGL.GLU import *
import Image
import fft.map.texture, fft.map.gns

from OpenGL.GL.ARB.shader_objects import *
from OpenGL.GL.ARB.fragment_shader import *
from OpenGL.GL.ARB.vertex_shader import *

try:
    from OpenGL.extensions import alternate
except ImportError:
    pass
else:
    glCreateShader = alternate( 'glCreateShader', glCreateShader, glCreateShaderObjectARB )
    glShaderSource = alternate( 'glShaderSource', glShaderSource, glShaderSourceARB)
    glCompileShader = alternate( 'glCompileShader', glCompileShader, glCompileShaderARB)
    glCreateProgram = alternate( 'glCreateProgram', glCreateProgram, glCreateProgramObjectARB)
    glAttachShader = alternate( 'glAttachShader', glAttachShader,glAttachObjectARB )
    glValidateProgram = alternate( 'glValidateProgram',glValidateProgram,glValidateProgramARB )
    glLinkProgram = alternate( 'glLinkProgram',glLinkProgram,glLinkProgramARB )
    glDeleteShader = alternate( 'glDeleteShader', glDeleteShader,glDeleteObjectARB )
    glUseProgram = alternate('glUseProgram',glUseProgram,glUseProgramObjectARB )

# Basic terrain slope types
flat = { 'ne': 0, 'se': 0, 'sw': 0, 'nw': 0 }
slant = { 'ne': 1, 'se': 0, 'sw': 0, 'nw': 1 }
convex = { 'ne': 1, 'se': 0, 'sw': 0, 'nw': 0 }
concave = { 'ne': 1, 'se': 1, 'sw': 0, 'nw': 1 }

# Terrain slope types at all the different orientations.
slope_types = {
    0x00: (flat, 0),
    0x85: (slant, 0),
    0x52: (slant, 90),
    0x25: (slant, 180),
    0x58: (slant, 270),
    0x41: (convex, 0),
    0x11: (convex, 90),
    0x14: (convex, 180),
    0x44: (convex, 270),
    0x96: (concave, 0),
    0x66: (concave, 90),
    0x69: (concave, 180),
    0x99: (concave, 270)
}

def compileShader( source, shaderType ):
        """Compile shader source of given type"""
        shader = glCreateShader(shaderType)
        glShaderSource( shader, source )
        glCompileShader( shader )
        return shader

def compileProgram(vertexSource=None, fragmentSource=None):
        program = glCreateProgram()

        if vertexSource:
                vertexShader = compileShader(
                        vertexSource, GL_VERTEX_SHADER_ARB
                )
                glAttachShader(program, vertexShader)
        if fragmentSource:
                fragmentShader = compileShader(
                        fragmentSource, GL_FRAGMENT_SHADER_ARB
                )
                glAttachShader(program, fragmentShader)

        glValidateProgram( program )
        glLinkProgram(program)

        try:
            if vertexShader:
                    glDeleteShader(vertexShader)
            if fragmentShader:
                    glDeleteShader(fragmentShader)
        except:
            pass

        return program

class Viewer():

    def __init__(self):
        # Display Lists
        self.mesh_dlists = []
        self.mesh_anim_dlists = []
        self.terrain_dlist = None
        # Camera settings
        self.use_ortho = True
        self.orthobase = 1
        self.magnify_one = None
        self.magnify = None
        self.fov = 27
        self.angle = 0
        self.angleinc = .05
        self.altitude = 30
        # View settings
        self.use_lights = True
        self.use_cutaway = True
        self.use_skirt = True
        self.use_terrain = False
        self.show_axes = True
        self.chroma_key = False
        # Animation settings
        self.anim_timer = 0
        self.do_animation = True
        # Texture settings
        self.use_texture = True
        self.textures = []
        # The map
        self.map = None
        self.map_is_loaded = False

    def load_gns(self, gns_filespec, index = 0):
        self.map_is_loaded = False
        self.map = fft.map.gns.File(gns_filespec)
        if type(index) == tuple:
            index = self.map.find_gns_index(index)
        self.map.get_resources(self.map.all_indexes[index])
        self.gl_create_textures()
        self.measure_map()
        self.set_orthobase()
        self.normalize_normals()
        self.gl_set_perspective()
        self.gl_create_lights()
        self.create_mesh_dlists()
        self.create_mesh_anim_dlists()
        self.create_terrain_dlist()
        self.map_is_loaded = True
        return index

    def load_traps(self, map_path, map_number):
        file_name = os.path.join(map_path, '..', 'BATTLE.BIN')
        F = fft.BattleBin(os.path.normpath(file_name))
        self.traps = F.read_trap(map_number)

    def load_placements(self, map_path, map_number, squads = None):
        file_name = os.path.join(map_path, '..', 'EVENT', 'ATTACK.OUT')
        F = fft.AttackOut(os.path.normpath(file_name))
        self.placements = F.read_placements(map_number)
        if not squads == None:
            keep_placements = []
            for place in self.placements:
                if (place['id'] == squads[0] or place['id'] == squads[1]):
                    keep_placements.append(place)
            self.placements = keep_placements

    def load_entds(self, map_path, id):
        F = fft.Entd(map_path, id)
        self.entds = F.read_entds()

    def gl_create_textures(self):
        if self.map.texture == None:
            self.gl_set_use_texture(False)
            return
        else:
            self.gl_set_use_texture(True)
        texFileName = self.map.texture
        # Create Texture
        if self.textures:
            glDeleteTextures(self.textures)
            self.textures = []
        newtex = glGenTextures(1)
        self.textures.append(newtex)
        glBindTexture(GL_TEXTURE_2D, newtex)
        image = Image.open(texFileName).convert("P")
        ix = image.size[0]
        iy = image.size[1]
        gray_palette = [(x / 3) * 17 for x in range(48)]
        # Palette loading, should probably go elsewhere
        texpalette = Image.new('RGBA', (256, 1));
        for p in range(16):
            for i in range(16):
                rgb = self.map.palettes_color[p][i*3:i*3+3]
                a = self.map.palettes_color_alpha[p][i * 3]
                texpalette.putpixel((p * 16 + i, 0), tuple(rgb + [a]))
        paletteimagestr = texpalette.tostring('raw', 'RGBA', 0, -1)
        newtexpalette = glGenTextures(1)
        self.textures.append(newtexpalette)
        glActiveTexture(GL_TEXTURE1)
        glBindTexture(GL_TEXTURE_1D, newtexpalette)
        glTexImage1D(GL_TEXTURE_1D, 0, GL_RGBA, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, paletteimagestr)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameterf(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glUniform1i(self.uniform_texpalette, 1)
        # End palette loading
        image.putpalette(gray_palette)
        #image.save("tex.png")
        glActiveTexture(GL_TEXTURE0)
        glBindTexture(GL_TEXTURE_2D, newtex)
        imagestr = image.convert('RGBA').tostring("raw", "RGBA", 0, -1)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB5_A1, ix, iy, 0, GL_RGBA, GL_UNSIGNED_BYTE, imagestr)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)

    def measure_map(self):
        max_dims = {'x': -32768, 'y': -32768, 'z': -32768}
        min_dims = {'x':  32767, 'y':  32767, 'z':  32767}
        for polygon in self.map.polygons:
            for point in 'abc':
                for dim in 'xyz':
                    max_dims[dim] = max(max_dims[dim], polygon[point][dim])
                    min_dims[dim] = min(min_dims[dim], polygon[point][dim])
        self.map.width = max_dims['x'] - min_dims['x']
        self.map.height = max_dims['y'] - min_dims['y']
        self.map.length = max_dims['z'] - min_dims['z']
        self.map.diameter = math.sqrt(self.map.width ** 2 + self.map.length ** 2) / 2

    def normalize_normals(self):
        for polygon in self.map.polygons:
            # Normalize normal vectors
            for point in 'abc':
                vertex = polygon[point]
                if not vertex["p"] == None:
                    vertex["p"] /= 4096.0
                    vertex["q"] /= 4096.0
                    vertex["r"] /= 4096.0

    def gl_create_lights(self):
        lights = self.map.lights
        diffuse_max = 4096.0
        ambient_max = 128.0
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, [lights[3]['r'] / ambient_max, lights[3]['g'] / ambient_max, lights[3]['b'] / ambient_max, 1.0])
        for i in range(3):
            light = lights[i]
            glLightID = GL_LIGHT0 + i
            glLightfv(glLightID, GL_DIFFUSE, [light['r'] / diffuse_max, light['g'] / diffuse_max, light['b'] / diffuse_max, 1.0])

    def gl_draw_background(self):
        glUseProgram(0)
        glBegin(GL_POLYGON)
        posz = -24000
        grad = self.map.background
        if self.use_ortho:
            posx = self.orthobase
            posy = self.orthobase / aspect
        else:
            factor = math.tan(math.radians(float(self.fov) / 2))
            posx = -posz * factor * aspect
            posy = -posz * factor
        if not self.chroma_key:
            glColor4f(grad[0]["r"] / 255, grad[0]["g"] / 255, grad[0]["b"] / 255, 1.0)
        else:
            glColor4f(1.0, 0.0, 1.0, 1.0)
        glVertex3f(-posx, posy, posz)
        glVertex3f(posx, posy, posz)
        if not self.chroma_key:
            glColor4f(grad[1]["r"] / 255, grad[1]["g"] / 255, grad[1]["b"] / 255, 1.0)
        glVertex3f(posx, -posy, posz)
        glVertex3f(-posx, -posy, posz)
        glEnd()
        glUseProgram(self.shader_program)

    def gl_draw_axes(self):
        global minDims
        glUseProgram(0)
        glDisable(GL_CULL_FACE)
        glColor3f(1, 0, 0)
        glBegin(GL_TRIANGLES)
        glVertex3i(  -5,   -5,     5)
        glVertex3i(1000,    0,     0)
        glVertex3i(   0,    0,     0)
        glColor3f(0, 1, 0)
        glVertex3i(  -5,   -5,    5)
        glVertex3i(   0, 1000,    0)
        glVertex3i(   0,    0,    0)
        glColor3f(0, 0, 1)
        glVertex3i(  -5,   -5,     5)
        glVertex3i(   0,    0, -1000)
        glVertex3i(   0,    0,     0)
        glEnd()
        glUseProgram(self.shader_program)
        glEnable(GL_CULL_FACE)

    def set_orthobase(self):
        if self.magnify == None:
            self.orthobase = self.map.diameter
        else:
            self.orthobase = self.magnify_one / (self.magnify / 100.0)
        self.gl_set_perspective()

    def set_azimuth_auto(self, event):
        if event.Checked():
            self.angleinc = 0.05
        else:
            self.angleinc = 0

    def set_azimuth(self, value):
        self.angle = value
        self.angleinc = 0

    def snap_azimuth(self, value):
        abs_value = abs(value)
        sign = value / abs_value
        if sign == 1: rounder = math.ceil
        elif sign == -1: rounder = math.floor
        factor = rounder((self.angle + sign * 0.001) / float(abs_value))
        self.set_azimuth(factor * abs_value)

    def set_altitude(self, value):
        self.altitude = value

    def get_magnify(self):
        return self.magnify_one / self.orthobase * 100.0

    def set_magnify(self, value):
        self.magnify = value
        self.set_orthobase()

    def set_fov(self, value):
        self.fov = value
        self.use_ortho = False
        self.gl_set_perspective()

    def set_fov_isometric(self, event):
        self.use_ortho = event.Checked()
        self.gl_set_perspective()

    def gl_draw_mesh(self, polygons, vis_angle = None):
        skirtPolys = []
        for poly in polygons:
            if self.use_cutaway and not vis_angle == None and not poly["vis"] == None:
                if poly["vis"][(7-vis_angle)+6] == 1:
                    continue
            glUniform1i(self.uniform_palette, int(poly['palette'] or 0))
            glBegin(GL_TRIANGLES)
            for point in ['a', 'b', 'c']:
                vertex = poly[point]
                if not vertex['u'] == None:
                    glNormal3f(vertex["p"], -vertex["q"], -vertex["r"])
                    if self.map.texture:
                        coords = self.get_gl_uv(poly, vertex)
                        glTexCoord2f(coords[0], coords[1])
                    else:
                        glColor4f(0.0, 0.0, 0.0, 1.0)
                    glVertex3i(vertex["x"], vertex["y"], -vertex["z"])
                else:
                    skirtPolys.append(vertex)
            glEnd()
        if self.use_skirt:
            glUseProgram(0)
            glColor4f(0.0, 0.0, 0.0, 1.0)
            glBegin(GL_TRIANGLES)
            for vertex in skirtPolys:
                glVertex3i(vertex["x"], vertex["y"], -vertex["z"])
            glEnd()
            glUseProgram(self.shader_program)

    def gl_initialize(self):
        glEnable(GL_CULL_FACE)
        glFrontFace(GL_CW)
        glEnable(GL_BLEND)
        glEnable(GL_ALPHA_TEST)
        glAlphaFunc(GL_GREATER, 0.5)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glClearColor(0.0, 0.0, 0.0, 0.0)    # Clear the Background to a Color
        glClearDepth(1.0)                   # Enables Clearing Of The Depth Buffer
        glDepthFunc(GL_LEQUAL)                # The Type Of Depth Test To Do
        glEnable(GL_DEPTH_TEST)             # Enables Depth Testing
        glShadeModel(GL_SMOOTH)             # Enables Smooth Color Shading
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        self.gl_create_shaders()

    def gl_create_shaders(self):
        self.shader_program = compileProgram(
[
        '''
        varying vec3 normal;
        void main() {
            normal = normalize(gl_NormalMatrix * gl_Normal);
            gl_TexCoord[0] = gl_MultiTexCoord0;
            gl_Position = ftransform();
        }
        '''
],
[
        '''
        varying vec3 normal;
        uniform sampler2D tex;
        uniform sampler1D texpalette;
        uniform int palette;
        uniform bool use_lights;
        uniform bool use_texture;
        void main() {
            vec3 ambient, color;
            vec4 texel, ct;
            float alpha;
            float index;

            if (use_texture) {
                texel = texture2D(tex,gl_TexCoord[0].st);
                index = texel.r * 15.0;
                index = sign(index) * floor(abs(index + 0.5));
                float subscript = float(palette) * 16.0 + index;
                ct = texture1D(texpalette, subscript / 255.0) * 8.225806;
                color = ct.rgb;
                alpha = ct.a;
            } else {
                ct = vec4(1.0);
                alpha = 1.0;
            }

            if (use_lights) {
                ambient = gl_LightModel.ambient.rgb;
                color = ambient * color;
                for (int i=0; i<3; i++) {
                    vec3 lightDir = gl_LightSource[i].position.xyz / 4096.0;
                    float intensity = max(dot(lightDir, normal), 0.0);
                    vec3 diffuse = min(gl_LightSource[i].diffuse.rgb, vec3(1.0)) * intensity;
                    diffuse = max(diffuse, vec3(0.0));
                    color = diffuse * ct.rgb * 2.0 + color;
                }
            }
            gl_FragColor = vec4(color, alpha);
        }
        ''',
]
)
        glUseProgram(self.shader_program)
        self.gl_initialize_uniforms()

    def gl_initialize_uniforms(self):
        self.uniform_palette = glGetUniformLocation(self.shader_program, 'palette')
        self.uniform_texpalette = glGetUniformLocation(self.shader_program, 'texpalette')
        self.uniform_use_lights = glGetUniformLocation(self.shader_program, 'use_lights')
        self.uniform_use_texture = glGetUniformLocation(self.shader_program, 'use_texture')
        glUniform1i(self.uniform_use_lights, self.use_lights)
        glUniform1i(self.uniform_use_texture, self.use_texture)

    def gl_set_use_lights(self, state):
        self.use_lights = bool(state)
        glUniform1i(self.uniform_use_lights, self.use_lights)

    def gl_set_use_texture(self, state):
        self.use_texture = bool(state) and bool(self.map.texture)
        glUniform1i(self.uniform_use_texture, self.use_texture)

    def gl_set_viewport(self, width, height):
        # The function called when our window is resized
        global W, H
        if height == 0: # Prevent A Divide By Zero If The Window Is Too Small
            height = 1
        W = width
        H = height
        # Reset The Current Viewport And Perspective Transformation
        glViewport(0, 0, W, H)
        self.magnify_one = W / 2.0
        if self.map_is_loaded:
            self.set_orthobase()
        self.gl_set_perspective()

    def create_terrain_dlist(self):
        tile_w = 28
        tile_h = 12
        if not self.terrain_dlist == None:
            glDeleteLists(self.terrain_dlist, 1)
        self.terrain_dlist = glGenLists(1)
        glNewList(self.terrain_dlist, GL_COMPILE)
        glUseProgram(0)
        glDisable(GL_CULL_FACE)
        for y in range(len(self.map.terrain)):
            level = self.map.terrain[y]
            for z in range(len(level)):
                row = level[z]
                for x in range(len(row)):
                    tile = row[x]
                    if tile['nocursor']:
                        continue
                    elif tile['nowalk']:
                        (r, g, b) = (1.0, 0.5, 0.5)
                    else:
                        (r, g, b) = (0.5, 0.5, 1.0)
                    draw_trap = False
                    for trap in self.traps:
                        if trap['x'] == x and trap['z'] == z and not tile['nowalk'] and y == 0:
                            draw_trap = True
                    draw_facing = None
                    for place in self.placements:
                        if place['z'] - 2 <= z <= place['z'] + 2 and place['x'] - 2 <= x <= place['x'] + 2:
                            relz = place['z'] - z + 2
                            relx = place['x'] - x + 2
                            if place['bitmap'][relz][relx] and not tile['nowalk'] and y == 0:
                                (r, g, b) = (0.5, 1.0, 0.5)
                                draw_facing = place['face'] + place['rotate']
                    for entd in self.entds:
                        if entd['x'] == x and entd['y'] == y and entd['z'] == z and not tile['nowalk']:
                            if entd['enemy']:
                                (r, g, b) = (1.0, 0.75, 0.25)
                            elif entd['guest']:
                                (r, g, b) = (0.0, 1.0, 0.875)
                            else:
                                (r, g, b) = (0.5, 1.0, 0.5)
                            draw_facing = (4 - entd['face']) + 1
                    color_mod = (x + z) % 2 * 0.05
                    glColor4f(r - color_mod, g - color_mod, b - color_mod, 1.0)
                    try:
                        (slope, rotation) = slope_types[tile['slopedir']]
                    except KeyError:
                        print 'Unknown slope type:', tile['slopedir']
                        glColor4f(1.0, 0.0, 0.0, 1.0)
                        (slope, rotation) = (flat, 0)
                    glPushMatrix()
                    glTranslate((x + 0.5) * tile_w,
                            tile_h * tile['height'] + 1,
                            -(z + 0.5) * tile_w)
                    if not draw_trap == False:
                        glPushMatrix()
                        glTranslate(0, (tile['slopeheight'] + tile['depth']) * tile_h + 1, 0)
                        glScalef(tile_w, tile_h, -tile_w)
                        glColor4f(1.0, 0.5, 1.0, 1.0)
                        glBegin(GL_TRIANGLE_FAN)
                        glVertex3f( 0.00, 1,  0.00)
                        glVertex3f(-0.25, 0, -0.25)
                        glVertex3f( 0.25, 0, -0.25)
                        glVertex3f( 0.25, 0,  0.25)
                        glVertex3f(-0.25, 0,  0.25)
                        glVertex3f(-0.25, 0, -0.25)
                        glEnd()
                        glColor4f(r - color_mod, g - color_mod, b - color_mod, 1.0)
                        glPopMatrix()
                    if not draw_facing == None:
                        glPushMatrix()
                        glTranslate(0, 1, 0)
                        glRotate(1 * 90, 0, 1, 0)
                        glRotate(draw_facing * 90, 0, 1, 0)
                        glScalef(tile_w, (tile['slopeheight'] + tile['depth']) * tile_h, -tile_w)
                        glColor4f(r / 2, g / 2, b / 2, 1.0)
                        glBegin(GL_TRIANGLES)
                        glVertex3f(-0.25, 1, -0.5)
                        glVertex3f( 0.25, 1, -0.5)
                        glVertex3f( 0.00, 1,  0.5)
                        glEnd()
                        glColor4f(r - color_mod, g - color_mod, b - color_mod, 1.0)
                        glPopMatrix()
                    glRotate(-rotation, 0, 1, 0)
                    glScalef(tile_w, tile_h, -tile_w)
                    glPushMatrix()
                    glScalef(1, tile['slopeheight'], 1)
                    glBegin(GL_TRIANGLE_FAN)
                    glVertex3f(-0.5, slope['sw'], -0.5)
                    glVertex3f( 0.5, slope['se'], -0.5)
                    glVertex3f( 0.5, slope['ne'],  0.5)
                    glVertex3f(-0.5, slope['nw'],  0.5)
                    glEnd()
                    glPopMatrix()
                    if tile['depth'] > 0:
                        # draw vertical quads to indicate water depth
                        glColor4f(r - color_mod - 0.2, g - color_mod - 0.2, b - color_mod - 0.2, 1.0)
                        glPushMatrix()
                        glScalef((tile_w + 1.0) / tile_w,
                                1,
                                (tile_w + 1.0) / tile_w)
                        glBegin(GL_QUAD_STRIP)
                        glVertex3f(-0.5, slope['sw'], -0.5)
                        glVertex3f(-0.5, slope['sw'] + tile['depth'], -0.5)
                        glVertex3f(-0.5, slope['nw'],  0.5)
                        glVertex3f(-0.5, slope['nw'] + tile['depth'],  0.5)
                        glVertex3f( 0.5, slope['ne'],  0.5)
                        glVertex3f( 0.5, slope['ne'] + tile['depth'],  0.5)
                        glVertex3f( 0.5, slope['se'], -0.5)
                        glVertex3f( 0.5, slope['se'] + tile['depth'], -0.5)
                        glVertex3f(-0.5, slope['sw'], -0.5)
                        glVertex3f(-0.5, slope['sw'] + tile['depth'], -0.5)
                        glEnd()
                        glPopMatrix()
                        glBegin(GL_TRIANGLE_FAN)
                        glVertex3f(-0.5, slope['sw'] + tile['depth'], -0.5)
                        glVertex3f( 0.5, slope['se'] + tile['depth'], -0.5)
                        glVertex3f( 0.5, slope['ne'] + tile['depth'],  0.5)
                        glVertex3f(-0.5, slope['nw'] + tile['depth'],  0.5)
                        glEnd()
                    glPopMatrix()
        glUseProgram(self.shader_program)
        glEnable(GL_CULL_FACE)
        glEndList()

    def create_mesh_dlists(self):
        range_num = 8
        if len(self.mesh_dlists):
            glDeleteLists(self.mesh_dlists[0], len(self.mesh_dlists))
            self.mesh_dlists = []
        if not self.use_cutaway:
            range_num = 1
        for i in range(range_num):
            self.mesh_dlists.append(glGenLists(1))
            glNewList(self.mesh_dlists[i], GL_COMPILE)
            self.gl_draw_mesh(self.map.polygons, i)
            glEndList()

    def create_mesh_anim_dlists(self):
        if len(self.mesh_anim_dlists):
            glDeleteLists(self.mesh_anim_dlists[0], len(self.mesh_anim_dlists))
            self.mesh_anim_dlists = []
        for i in range(len(self.map.polygons_extra)):
            self.mesh_anim_dlists.append(glGenLists(1))
            glNewList(self.mesh_anim_dlists[i], GL_COMPILE)
            self.gl_draw_mesh(self.map.polygons_extra[i])
            glEndList()

    def gl_draw(self):
        # The main drawing function.

        # Clear The Screen And The Depth Buffer
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()                    # Reset The View

        self.gl_draw_background()

        # Move camera into position
        glTranslate(0, 0, -self.orthobase * 3)
        glRotate(self.altitude, 1, 0, 0)
        glRotate(self.angle, 0, 1, 0)

        # Move map to center of camera's vision
        glTranslate(-self.map.width / 2, -self.map.height / 3, self.map.length / 2)

        if self.show_axes:
            self.gl_draw_axes()

        if self.map.texture:
            glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, [1.0, 1.0, 1.0, 1.0])

        if self.use_terrain:
            glCallList(self.terrain_dlist)

        if self.use_lights:
            for i in range(3):
                glLightID = GL_LIGHT0 + i
                light = self.map.lights[i]
                glLightiv(glLightID, GL_POSITION, [light['x'], light['y'], -light['z'], 1])

        range_num = 8
        if not self.use_cutaway:
            range_num = 1
        for i in range(range_num):
            if (7-i)*45 <= self.angle < (7-i)*45 + 45 or not self.use_cutaway:
                glCallList(self.mesh_dlists[i])

        for i in range(len(self.mesh_anim_dlists)):
            glPushMatrix()
            for t in self.map.polygons_extra_transform[i]:
                glTranslate(t['trans'][0], t['trans'][1], -t['trans'][2])
                glScalef(t['scale'][0], t['scale'][1], t['scale'][2])
                if not t['angle'] == None:
                    glRotatef(t['angle'] * self.anim_timer, t['axis'][0] , t['axis'][1] , t['axis'][2])
            glCallList(self.mesh_anim_dlists[i])
            glPopMatrix()

        self.angle += self.angleinc
        self.mod_angles()

        # Animate the scene
        if self.do_animation:
            self.anim_timer += 0.5
            self.anim_timer %= 360

    def gl_set_perspective(self):
        global aspect
        # Reset The Projection Matrix
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        # Set orthographic or perspective projection
        aspect = float(W) / float (H)
        if self.use_ortho:
            glOrtho(-self.orthobase, self.orthobase, -self.orthobase / aspect,
                    self.orthobase / aspect, 0.1, 24000)
        else:
            gluPerspective(self.fov, aspect, 0.1, 24000.0)
        glMatrixMode(GL_MODELVIEW)

    def get_gl_uv(self, poly, vertex):
        x_dim = 255.0
        y_dim = 1023.0
        page = poly['page']
        if page == None:
            page = 0
        coords = [
                vertex['u'] / x_dim,
                (y_dim - ((page & 0x03) * 256 + vertex['v'])) / y_dim
        ]
        return coords

    def mod_angles(self):
        self.angle %= 360
        self.altitude %= 360
